@@ -1,10 +1,14 @@
-"""Automatic bugs filler"""
+"""Automatic bugs filer"""
 
+import contextlib
 import json
+import sys
 import urllib.request as urllib
 from collections import defaultdict
 from functools import partial
 from itertools import chain
+from pathlib import Path
+from urllib.parse import urlencode
 
 from pkgcheck import const as pkgcheck_const
 from pkgcheck.addons import ArchesAddon, init_addon
@@ -13,11 +17,12 @@ from pkgcheck.checks import visibility
 from pkgcheck.scripts import argparse_actions
 from pkgcore.ebuild.atom import atom
 from pkgcore.ebuild.ebuild_src import package
+from pkgcore.ebuild.errors import MalformedAtom
 from pkgcore.ebuild.misc import sort_keywords
 from pkgcore.repository import multiplex
 from pkgcore.restrictions import boolean, packages, values
 from pkgcore.test.misc import FakePkg
-from pkgcore.util import commandline
+from pkgcore.util import commandline, parserestrict
 from snakeoil.cli import arghparse
 from snakeoil.cli.input import userquery
 from snakeoil.formatters import Formatter
@@ -37,16 +42,17 @@ bugs.add_argument(
     metavar="KEY",
     help="Bugzilla API key",
     docs="""
-        The Bugzilla API key to use for authentication. Used mainly to overcome
-        rate limiting done by bugzilla server. This tool doesn't perform any
-        bug editing, just fetching info for the bug.
+        The Bugzilla API key to use for authentication. WARNING: using this
+        option will expose your API key to other users of the same system.
+        Consider instead saving your API key in a file named ~/.bugz_token.
     """,
 )
 bugs.add_argument(
     "targets",
     metavar="target",
-    nargs="+",
+    nargs="*",
     action=commandline.StoreTarget,
+    use_sets="sets",
     help="extended atom matching of packages",
 )
 bugs.add_argument(
@@ -64,6 +70,18 @@ bugs.add_argument(
         package is maintainer-needed, always add CC-ARCHES.
     """,
 )
+bugs.add_argument(
+    "--blocks",
+    metavar="BUG",
+    action=arghparse.CommaSeparatedValuesAppend,
+    default=[],
+    help="bugs which should be blocked by newly created bugs",
+    docs="""
+        Collection of bug ids which should be blocked by newly created bugs.
+        Only bugs created for passed targets would be blockers, excluding other
+        bugs which were created as dependencies.
+    """,
+)
 
 bugs.add_argument(
     "--cache",
@@ -127,6 +145,16 @@ def _get_suggested_keywords(repo, pkg: package):
     return frozenset({x for x in match_keywords if "-" not in x})
 
 
+def parse_atom(pkg: str):
+    try:
+        return atom(pkg)
+    except MalformedAtom as exc:
+        try:
+            return atom(f"={pkg}")
+        except MalformedAtom:
+            raise exc
+
+
 class GraphNode:
     __slots__ = ("pkgs", "edges", "bugno")
 
@@ -170,12 +198,14 @@ class GraphNode:
                 keywords.clear()
                 keywords.add("*")
 
-    def file_bug(self, api_key: str, auto_cc_arches: frozenset[str], observer=None) -> int:
+    def file_bug(
+        self, api_key: str, auto_cc_arches: frozenset[str], block_bugs: list[int], observer=None
+    ) -> int:
         if self.bugno is not None:
             return self.bugno
         for dep in self.edges:
             if dep.bugno is None:
-                dep.file_bug(api_key, auto_cc_arches, observer)
+                dep.file_bug(api_key, auto_cc_arches, (), observer)
         maintainers = dict.fromkeys(
             maintainer.email for pkg, _ in self.pkgs for maintainer in pkg.maintainers
         )
@@ -185,19 +215,24 @@ class GraphNode:
             keywords = []
         maintainers = tuple(maintainers) or ("maintainer-needed@gentoo.org",)
 
+        summary = f"{', '.join(pkg.versioned_atom.cpvstr for pkg, _ in self.pkgs)}: stablereq"
+        if len(summary) > 90 and len(self.pkgs) > 1:
+            summary = f"{self.pkgs[0][0].versioned_atom.cpvstr} and friends: stablereq"
+
         request_data = dict(
             Bugzilla_api_key=api_key,
             product="Gentoo Linux",
             component="Stabilization",
             severity="enhancement",
             version="unspecified",
-            summary=f"{', '.join(pkg.versioned_atom.cpvstr for pkg, _ in self.pkgs)}: stablereq",
+            summary=summary,
             description="Please stabilize",
             keywords=keywords,
             cf_stabilisation_atoms="\n".join(self.lines()),
             assigned_to=maintainers[0],
             cc=maintainers[1:],
             depends_on=list({dep.bugno for dep in self.edges}),
+            blocks=block_bugs,
         )
         request = urllib.Request(
             url="https://bugs.gentoo.org/rest/bug",
@@ -238,7 +273,7 @@ class DependencyGraph:
 
     def find_best_match(self, restrict, pkgset: list[package]) -> package:
         restrict = boolean.AndRestriction(
-            restrict,
+            *restrict,
             packages.PackageRestriction("properties", values.ContainmentMatch("live", negate=True)),
         )
         # prefer using already selected packages in graph
@@ -251,6 +286,16 @@ class DependencyGraph:
                 return match
         return matches[0]
 
+    def extend_targets_stable_groups(self, groups):
+        stabilization_groups = self.options.repo.stabilization_groups
+        search_repo = self.options.search_repo
+        for group in groups:
+            for pkg in stabilization_groups[group]:
+                try:
+                    yield None, self.find_best_match([pkg], search_repo.match(pkg)).versioned_atom
+                except (ValueError, IndexError):
+                    self.err.write(f"Unable to find match for {pkg.unversioned_atom}")
+
     def _find_dependencies(self, pkg: package, keywords: set[str]):
         check = visibility.VisibilityCheck(self.options, profile_addon=self.profile_addon)
 
@@ -264,13 +309,19 @@ class DependencyGraph:
         for pkgname, problems in issues.items():
             pkgset: list[package] = self.options.repo.match(atom(pkgname))
             try:
-                combined = boolean.AndRestriction(*set().union(*problems.values()))
-                match = self.find_best_match(combined, pkgset)
+                match = self.find_best_match(set().union(*problems.values()), pkgset)
                 yield match, set(problems.keys())
-            except ValueError:
+            except (ValueError, IndexError):
                 results: dict[package, set[str]] = defaultdict(set)
                 for keyword, deps in problems.items():
-                    match = self.find_best_match(deps, pkgset)
+                    try:
+                        match = self.find_best_match(deps, pkgset)
+                    except (ValueError, IndexError):
+                        deps_str = " , ".join(map(str, deps))
+                        self.err.error(
+                            f"unable to find match for restrictions: {deps_str}",
+                        )
+                        raise
                     results[match].add(keyword)
                 yield from results.items()
 
@@ -285,7 +336,11 @@ class DependencyGraph:
                 vertices[pkg].pkgs[0][1].update(keywords)
                 continue
 
+            pkg_has_stable = any(x[0] not in "-~" for x in pkg.keywords)
             keywords.update(_get_suggested_keywords(self.options.repo, pkg))
+            if pkg_has_stable and not keywords:  # package already done
+                self.out.write(f"Nothing to stable for {pkg.unversioned_atom}")
+                continue
             assert (
                 keywords
             ), f"no keywords for {pkg.versioned_atom}, currently unsupported by tool: https://github.com/pkgcore/pkgdev/issues/123"
@@ -302,7 +357,9 @@ class DependencyGraph:
 
         for src, dst in edges:
             vertices[src].edges.add(vertices[dst])
-        self.starting_nodes = {vertices[starting_node] for starting_node in targets}
+        self.starting_nodes = {
+            vertices[starting_node] for starting_node in targets if starting_node in vertices
+        }
 
     def output_dot(self, dot_file):
         with open(dot_file, "w") as dot:
@@ -310,6 +367,8 @@ class DependencyGraph:
             dot.write("\trankdir=LR;\n")
             for node in self.nodes:
                 node_text = "\\n".join(node.lines())
+                if node.bugno is not None:
+                    node_text += f"\\nbug #{node.bugno}"
                 dot.write(f'\t{node.dot_edge}[label="{node_text}"];\n')
                 for other in node.edges:
                     dot.write(f"\t{node.dot_edge} -> {other.dot_edge};\n")
@@ -318,6 +377,7 @@ class DependencyGraph:
 
     def merge_nodes(self, nodes: tuple[GraphNode, ...]) -> GraphNode:
         self.nodes.difference_update(nodes)
+        is_start = bool(self.starting_nodes.intersection(nodes))
         self.starting_nodes.difference_update(nodes)
         new_node = GraphNode(list(chain.from_iterable(n.pkgs for n in nodes)))
 
@@ -330,6 +390,8 @@ class DependencyGraph:
                 node.edges.add(new_node)
 
         self.nodes.add(new_node)
+        if is_start:
+            self.starting_nodes.add(new_node)
         return new_node
 
     @staticmethod
@@ -345,17 +407,15 @@ class DependencyGraph:
         return ()
 
     def merge_cycles(self):
-        new_starts = set()
-        while self.starting_nodes:
-            starting_node = self.starting_nodes.pop()
+        start_nodes = set(self.starting_nodes)
+        while start_nodes:
+            starting_node = start_nodes.pop()
             assert starting_node in self.nodes
             while cycle := self._find_cycles(tuple(self.nodes), [starting_node]):
-                print("Found cycle:", " -> ".join(str(n) for n in cycle))
+                self.out.write("Found cycle: ", " -> ".join(str(n) for n in cycle))
                 new_node = self.merge_nodes(cycle)
                 if starting_node not in self.nodes:
                     starting_node = new_node
-            new_starts.add(starting_node)
-        self.starting_nodes.update(new_starts)
 
     def merge_new_keywords_children(self):
         repo = self.options.search_repo
@@ -372,50 +432,142 @@ class DependencyGraph:
                 existing_keywords = frozenset().union(
                     *(
                         pkgver.keywords
-                        for pkg in node.pkgs
-                        for pkgver in repo.match(pkg[0].unversioned_atom)
+                        for pkg, _ in node.pkgs
+                        for pkgver in repo.match(pkg.unversioned_atom)
                     )
                 )
                 if existing_keywords & frozenset().union(*(pkg[1] for pkg in node.pkgs)):
                     continue  # not fully new keywords
                 orig = next(iter(origs))
-                print(f"Merging {node} into {orig}")
+                self.out.write(f"Merging {node} into {orig}")
                 self.merge_nodes((orig, node))
                 found_someone = True
                 break
 
-    def file_bugs(self, api_key: str, auto_cc_arches: frozenset[str]):
+    def merge_stabilization_groups(self):
+        for group, pkgs in self.options.repo.stabilization_groups.items():
+            restrict = packages.OrRestriction(*pkgs)
+            mergable = tuple(
+                node for node in self.nodes if any(restrict.match(pkg) for pkg, _ in node.pkgs)
+            )
+            if mergable:
+                self.out.write(f"Merging @{group} group nodes: {mergable}")
+                self.merge_nodes(mergable)
+
+    def scan_existing_bugs(self, api_key: str):
+        params = urlencode(
+            {
+                "Bugzilla_api_key": api_key,
+                "include_fields": "id,cf_stabilisation_atoms",
+                "component": "Stabilization",
+                "resolution": "---",
+                "f1": "cf_stabilisation_atoms",
+                "o1": "anywords",
+                "v1": {pkg[0].unversioned_atom for node in self.nodes for pkg in node.pkgs},
+            },
+            doseq=True,
+        )
+        request = urllib.Request(
+            url="https://bugs.gentoo.org/rest/bug?" + params,
+            method="GET",
+            headers={
+                "Content-Type": "application/json",
+                "Accept": "application/json",
+            },
+        )
+        with urllib.urlopen(request, timeout=30) as response:
+            reply = json.loads(response.read().decode("utf-8"))
+        for bug in reply["bugs"]:
+            bug_atoms = (
+                parse_atom(line.split(" ", 1)[0]).unversioned_atom
+                for line in map(str.strip, bug["cf_stabilisation_atoms"].splitlines())
+                if line
+            )
+            bug_match = boolean.OrRestriction(*bug_atoms)
+            for node in self.nodes:
+                if node.bugno is None and all(bug_match.match(pkg[0]) for pkg in node.pkgs):
+                    node.bugno = bug["id"]
+                    self.out.write(
+                        self.out.fg("yellow"),
+                        f"Found https://bugs.gentoo.org/{node.bugno} for node {node}",
+                        self.out.reset,
+                    )
+                    break
+
+    def file_bugs(self, api_key: str, auto_cc_arches: frozenset[str], block_bugs: list[int]):
         def observe(node: GraphNode):
             self.out.write(
                 f"https://bugs.gentoo.org/{node.bugno} ",
                 " | ".join(node.lines()),
                 " depends on bugs ",
-                {dep.bugno for dep in node.edges},
+                {dep.bugno for dep in node.edges} or "{}",
             )
             self.out.flush()
 
         for node in self.starting_nodes:
-            node.file_bug(api_key, auto_cc_arches, observe)
+            node.file_bug(api_key, auto_cc_arches, block_bugs, observe)
+
+
+def _load_from_stdin(out: Formatter, err: Formatter):
+    if not sys.stdin.isatty():
+        out.warn("No packages were specified, reading from stdin...")
+        for line in sys.stdin.readlines():
+            if line := line.split("#", 1)[0].strip():
+                yield line, parserestrict.parse_match(line)
+        # reassign stdin to allow interactivity (currently only works for unix)
+        sys.stdin = open("/dev/tty")
+    else:
+        raise arghparse.ArgumentError(None, "reading from stdin is only valid when piping data in")
+
+
+def _parse_targets(search_repo, targets):
+    for _, target in targets:
+        try:
+            yield max(search_repo.itermatch(target))
+        except ValueError:
+            raise ValueError(f"Restriction {target} has no match in repository")
 
 
 @bugs.bind_main_func
 def main(options, out: Formatter, err: Formatter):
     search_repo = options.search_repo
-    targets = [max(search_repo.itermatch(target)) for _, target in options.targets]
+    options.targets = options.targets or []
     d = DependencyGraph(out, err, options)
+    options.targets.extend(d.extend_targets_stable_groups(options.sets or ()))
+    if not options.targets:
+        options.targets = list(_load_from_stdin(out, err))
+    targets = list(_parse_targets(search_repo, options.targets))
     d.build_full_graph(targets)
     d.merge_cycles()
     d.merge_new_keywords_children()
+    d.merge_stabilization_groups()
 
     for node in d.nodes:
         node.cleanup_keywords(search_repo)
 
+    if options.api_key is None:
+        bugz_token_file = Path.home() / ".bugz_token"
+        if bugz_token_file.is_file:
+            options.api_key = bugz_token_file.read_text().strip()
+
+    if not d.nodes:
+        out.write(out.fg("red"), "Nothing to do, exiting", out.reset)
+        return 1
+
+    if userquery("Check for open bugs matching current graph?", out, err, default_answer=False):
+        d.scan_existing_bugs(options.api_key)
+
     if options.dot is not None:
         d.output_dot(options.dot)
         out.write(out.fg("green"), f"Dot file written to {options.dot}", out.reset)
 
+    bugs_count = len(tuple(node for node in d.nodes if node.bugno is None))
+    if bugs_count == 0:
+        out.write(out.fg("red"), "Nothing to do, exiting", out.reset)
+        return 1
+
     if not userquery(
-        f"Continue and create {len(d.nodes)} stablereq bugs?", out, err, default_answer=False
+        f"Continue and create {bugs_count} stablereq bugs?", out, err, default_answer=False
     ):
         return 1
 
@@ -424,4 +576,5 @@ def main(options, out: Formatter, err: Formatter):
         return 1
 
     disabled, enabled = options.auto_cc_arches
-    d.file_bugs(options.api_key, frozenset(enabled).difference(disabled))
+    blocks = list(frozenset(map(int, options.blocks)))
+    d.file_bugs(options.api_key, frozenset(enabled).difference(disabled), blocks)
